# -*- coding: utf-8 -*-
# Copyright 2009-2014, Red Hat, Inc.
# License: GPL-2.0+ <http://spdx.org/licenses/GPL-2.0+>
# See the LICENSE file for more details on Licensing

from __future__ import absolute_import

DOCUMENTATION = """
module: resultsdb_directive
short_description: send task results to ResultsDB or check YAML correctness
description: |
  Send check output to `ResultsDB <https://fedoraproject.org/wiki/ResultsDB>`_.

  If reporting to ResultsDB is disabled in the config (``report_to_resultsdb``
  option, disabled by default for the development profile), the directive at
  least checks whether the check output is in valid :ref:`resultyaml-format` and logs the
  details that would have been reported. For this reason, it's useful to use
  resultsdb directive always, even when you don't have any ResultsDB server
  configured.

parameters:
  results:
    required: true
    description: Variable containing the results of the check in the ResultYAML
                 format. You must use either ``results`` or ``file`` parameter.
    type: str
  file:
    required: true
    description: Path to a file containing the results of the check in the
                 ResultYAML format.
                 You must use either ``results`` or ``file`` parameter.
    type: str
returns: |
  YAML formatted output as taken from input, enhanced with result-id's for each
  successfully stored result. The result-id information then can be used by
  other directives. If reporting to ResultsDB is disabled, the returned YAML
  does not contain this extra field.
raises: |
  * :class:`.TaskotronDirectiveError`: if ``results`` is not in a valid YAML
    format, or when  ``item`` or ``type`` are missing in the YAML's result data,
    or when results are not allowed to be posted into a given namespace, or
    the namespace does not exist, or if both ``results`` and ``file`` parameters
    are used;
  * :class:`resultsdb_api.ResultsDBapiException`: for any errors coming from
    the ResultsDB server
version_added: 0.4
"""

EXAMPLES = """
These two actions first run ``run_rpmlint.py`` and export its YAML output to
``${rpmlint_output}``, and then feed this YAML output to ``resultsdb``
directive::

    - name: run rpmlint on downloaded rpms
      python:
          file: run_rpmlint.py
          callable: run
          workdir: ${workdir}
      export: rpmlint_output

    - name: report results to resultsdb
      resultsdb:
          results: ${rpmlint_output}

This example shows using the ``file`` parameter. The ``randomtest`` produces
a ``testoutput.yml`` file containing results in the YAML format, that the
``resultsdb`` directive then uses::

    - name: run randomtest on downloaded rpms
      shell:
          - bash ./randomtest.sh --outfile=${workdir}/testoutput.yml

    - name report results to resultsdb
      resultsdb:
          file: ${workdir}/testoutput.yml

If ResultsDB reporting is configured in the config file, it will get saved on
the ResultsDB server, otherwise only YAML compliance will get checked and some
summary will be printed out into the log, like this::

    [libtaskotron:resultsdb_directive.py:143] 2014-06-24 13:55:27 INFO    YAML is OK.
    [libtaskotron:resultsdb_directive.py:144] 2014-06-24 13:55:27 INFO    Reporting to ResultsDB is disabled.
    [libtaskotron:resultsdb_directive.py:145] 2014-06-24 13:55:27 INFO    Once enabled, the following would be reported:
    [libtaskotron:resultsdb_directive.py:146] 2014-06-24 13:55:27 INFO    <CheckDetail: {'_outcome': 'PASSED',
     'item': 'xchat-2.8.8-21.fc20.x86_64.rpm',
     'keyvals': {},
     'output': '<stripped out>',
     'report_type': 'koji_build',
     'summary': 'RPMLINT PASSED for xchat-2.8.8-21.fc20.x86_64.rpm'}>
    <CheckDetail: {'_outcome': 'PASSED',
     'item': 'xchat-tcl-2.8.8-21.fc20.x86_64.rpm',
     'keyvals': {},
     'output': '<stripped out>',
     'report_type': 'koji_build',
     'summary': 'RPMLINT PASSED for xchat-tcl-2.8.8-21.fc20.x86_64.rpm'}>
"""

import os
import configparser

from libtaskotron.directives import BaseDirective

from libtaskotron import check
from libtaskotron import config
from libtaskotron import buildbot_utils
from libtaskotron.exceptions import TaskotronDirectiveError, TaskotronValueError
from libtaskotron.logger import log
from libtaskotron.ext.fedora import rpm_utils

import resultsdb_api

directive_class = 'ResultsdbDirective'

class ResultsdbDirective(BaseDirective):

    def __init__(self, resultsdb = None):
        self.resultsdb = resultsdb

        conf = config.get_config()
        self.masterurl = conf.taskotron_master
        self.task_stepname = conf.buildbot_task_step
        self.execdb_server = "%s/jobs" % conf.execdb_server
        self.artifacts_baseurl = conf.artifacts_baseurl

        if self.resultsdb is None:
            self.resultsdb = resultsdb_api.ResultsDBapi(conf.resultsdb_server)

        self._ensured_testcases = []

    def ensure_testcase_exists(self, name):
        """ Make sure that the testcase exists in resultsdb, otherwise create
        the testcase using a dummy url as a reference

        :param str name: name of the testcase to check for
        """

        if name in self._ensured_testcases:
            return

        try:
            self.resultsdb.get_testcase(name)
            self._ensured_testcases.append(name)
            return
        except resultsdb_api.ResultsDBapiException, e:
            if not e.message.startswith('Testcase not found'):
                raise e

        # since the testcase doesn't exist, create it with a dummy value for url
        # it can be updated later when it's not blocking results reporting
        dummyurl = 'http://faketestcasesRus.com/%s' % name
        self.resultsdb.create_testcase(name, dummyurl)
        self._ensured_testcases.append(name)

    def create_resultsdb_job(self, name, uuid=None, refurl=None):
        """ Create a job in resultsdb for reporting results against

        :param str name: name of job to report against
        :param str uuid: UUID of the job (most probably provided by ExecDB)
        :param str refurl: url pointing to the execution overview.
           If set to None, ExecDB url is created from UUID
        :returns: dict of rendered json results
        """

        url = "%s/%s" % (self.execdb_server, uuid)
        if refurl is not None:
            url = refurl

        jobdata = self.resultsdb.create_job(url, status='SCHEDULED', name=name,
                                            uuid=uuid)
        self.resultsdb.update_job(id=jobdata['id'], status='RUNNING')

        return jobdata

    def complete_resultsdb_job(self, jobid):
        """ Change the resultsdb job to a status of ``COMPLETED``, indicating
        that the reporting is complete.

        :param int jobid: The resultsdb jobid to be modified
        :returns: json representation of returned data from the resultsdb instance
        """
        return self.resultsdb.update_job(jobid, status='COMPLETED')

    def get_artifact_path(self, artifactsdir, artifact):
        """Return the relative path of :attr str artifact: inside the
        :attr str artifactsdir:.
        :returns: relative path to the artifact file or None, if the file
                  does not exist, or is outside the artifactsdir.
        """
        artifactsdir = os.path.realpath(artifactsdir)
        if os.path.isabs(artifact):
            artifact_path = artifact
        else:
            artifact_path = os.path.join(artifactsdir, artifact)
        artifact_path = os.path.realpath(artifact_path)

        if not os.path.exists(artifact_path):
            log.warn('Artifact %r does not exist, ignoring' % artifact_path)
            return None
        elif not artifact_path.startswith(artifactsdir):
            log.warn('Artifact %r is placed outside of artifacts directory %r, ignoring',
                     artifact_path, artifactsdir)
            return None

        return os.path.relpath(artifact_path, start=artifactsdir)

    def check_namespace(self, checkname, arg_data):
        conf_ns = config.load_namespaces_config()
        ns_repos = None

        # check if the namespace exists
        for ns in conf_ns['namespaces_safe']:
            if checkname.startswith(ns+'.'):
                log.debug('Namespace %s found in the safe namespaces.', ns)
                return
        else:
            for ns, rs in conf_ns['namespaces_whitelist'].iteritems():
                if checkname.startswith(ns+'.'):
                    log.debug('Namespace %s found in the namespace whitelist.', ns)
                    ns_repos = rs
                    break
            else:
                raise TaskotronDirectiveError('Namespace %s does not exist.')

        try:
            taskdir = os.path.dirname(os.path.abspath(arg_data['task']))
            gitconfig_filename = os.path.join(taskdir, '.git/config')

            gitconfig = configparser.ConfigParser()
            gitconfig.read(gitconfig_filename)
            task_repo_url = gitconfig['remote "origin"']['url']
        except TypeError, e:
            log.exception(e)
            raise TaskotronDirectiveError("Could not find task's git remote 'origin' url"
                                          "in %s" % gitconfig_filename)

        try:
            if not [ns_repo for ns_repo in ns_repos if task_repo_url.strip().startswith(ns_repo)]:
                raise TaskotronDirectiveError

            log.debug('Repo %s found in the whitelist', task_repo_url)
            # dist-git style tasks
            # here we need to not only check the root namespace but also second level ns,
            # e.g. a firefox check can post results into pkg.firefox
            if checkname.startswith('pkg.'):
                pkg_name = rpm_utils.rpmformat(arg_data['item'], fmt='n')
                pkg_ns = 'pkg.%s.' % pkg_name
                if not checkname.startswith(pkg_ns):
                    raise TaskotronDirectiveError
        except TaskotronDirectiveError:
            raise TaskotronDirectiveError("%s is not allowed to post results into %s "
                                          "namespace. Not posting results." %
                                          (arg_data['checkname'], arg_data['namespace']))

    def process(self, params, arg_data):
        # checking if reporting is enabled is done after importing yaml which
        # serves as validation of input results

        if 'file' in params and 'results' in params:
            raise TaskotronDirectiveError("Either `file` or `results` can be used, not both.")

        try:
            if params.get('file', None):
                with open(params['file']) as resultfile:
                    params['results'] = resultfile.read()

            check_details = check.import_YAML(params['results'])
            log.debug("YAML output parsed OK.")
        except (TaskotronValueError, IOError) as e:
            raise TaskotronDirectiveError("Failed to load results: %s" % e.message)

        for detail in check_details:
            if not (detail.item and detail.report_type):
                raise TaskotronDirectiveError(
                    "The resultsdb directive requires "
                    "'item' and 'type' to be present in the YAML data.")

        conf = config.get_config()
        if not conf.report_to_resultsdb:
            log.info("Reporting to ResultsDB is disabled. Once enabled, the "
                     "following would get reported:\n%s" % params['results'])
            return check.export_YAML(check_details)

        checkname = '%s.%s' % (arg_data['namespace'], arg_data['checkname'])

        # find out if the task is allowed to post results into the namespace
        if config.get_config().profile == config.ProfileName.PRODUCTION:
            self.check_namespace(checkname, arg_data)

        # for now, we're creating the resultsdb job at reporting time
        # the job needs to be 'RUNNING' in order to append any results
        job_url, log_url = buildbot_utils.get_urls(arg_data['jobid'],
                                                   self.masterurl,
                                                   self.task_stepname)

        job_data = self.create_resultsdb_job(checkname, uuid=arg_data['uuid'])

        log.info('Posting %s results to ResultsDB...' % len(check_details))
        for detail in check_details:
            checkname = '%s.%s' % (arg_data['namespace'],
                                   detail.checkname or arg_data['checkname'])

            self.ensure_testcase_exists(checkname)
            result_log_url = log_url
            if detail.artifact:
                artifact_path = self.get_artifact_path(
                    arg_data['artifactsdir'],
                    detail.artifact
                    )
                if artifact_path:
                    result_log_url = "%s/all/%s/task_output/%s" % (
                        self.artifacts_baseurl,
                        arg_data['uuid'],
                        artifact_path,
                        )
            try:
                result = self.resultsdb.create_result(
                    job_id=job_data['id'],
                    testcase_name=checkname,
                    outcome=detail.outcome,
                    summary=detail.note or None,
                    log_url=result_log_url,
                    item=detail.item,
                    type=detail.report_type,
                    **detail.keyvals)
                detail._internal['resultsdb_result_id'] = result['id']

            except resultsdb_api.ResultsDBapiException, e:
                log.error(e)
                log.error("Failed to store ResulsDB: `%s` `%s` `%s`",
                           detail.item, checkname, detail.outcome)

        # move job to completed
        self.complete_resultsdb_job(job_data['id'])

        return check.export_YAML(check_details)
