# -*- coding: utf-8 -*-
# Copyright 2009-2014, Red Hat, Inc.
# License: GPL-2.0+ <http://spdx.org/licenses/GPL-2.0+>
# See the LICENSE file for more details on Licensing

from __future__ import absolute_import

import xunitparser

from libtaskotron.directives import BaseDirective
from libtaskotron.exceptions import TaskotronDirectiveError
from libtaskotron.check import CheckDetail, export_YAML

DOCUMENTATION = """
module: xunit_directive
short_description: parse XML xUnit file
description: |
  Parse test ouput in XML xUnit format and generate ResultYAML for further use (i.e. in resultsdb)
parameters:
  file:
    required: true
    description: xml file to be parsed, relative path is to the task formula file
    type: str
  aggregation:
    required: false
    description: type of result aggregation, ``none`` means no aggregation will be use and all
      testcases will be reporter as separate results. ``allpass`` will agregate testcases into
      one results with outcome ``PASSED`` if all testcases passed otherwise ``FAILED`` outcome is
      set.
      If not provided ``allpass`` is default value.
    type: str
returns: |
  Returns output of the check in the ResultYAML format directly usable in
  :ref:`resultsdb directive <resultsdb_directive>`

raises: |
  * :class:`.TaskotronDirectiveError`: if ``file`` was not provided or aggregation
    was set to unknown type.
version_added: 0.4.14
"""

EXAMPLES = """
This example parses XML file located in ``artifactsdir`` generates ResultYAML with no aggregation
and exports it to variable which is finally passed to ``resultsdb``::

  - name: parse test output
    xunit:
        file: ${artifactsdir}/output.xml
        aggregation: none
    export: test_output

  - name: report results to resultsdb
    resultsdb:
        results: ${test_output}
"""

aggregations = ['none', 'allpass']

directive_class = 'XunitDirective'


class XunitDirective(BaseDirective):
    def process(self, params, arg_data):
        try:
            artifact = params['file']
        except KeyError:
            raise TaskotronDirectiveError('No file provided.')

        aggregation = params.get('aggregation', 'allpass')

        if not arg_data['item'] or not arg_data['type']:
            raise TaskotronDirectiveError('Item and its type has to be provided.')

        if aggregation not in aggregations:
            raise TaskotronDirectiveError(
                    'Aggregation must be one of %s.' % ', '.join(aggregations))

        with open(artifact) as xmlfile:
            testsuite, testresult = xunitparser.parse(xmlfile)

        if aggregation == 'none':
            details = []
            for tc in testsuite:
                outcome = 'PASSED' if tc.good else 'FAILED'
                details.append(CheckDetail(item=arg_data['item'],
                                           checkname="%s.%s" % (arg_data['checkname'],
                                                                tc.methodname),
                                           report_type=arg_data['type'],
                                           outcome=outcome))

            return export_YAML(details)
        elif aggregation == 'allpass':
            summary = "%d PASSED, %d FAILED"
            passed = 0
            failed = 0
            final_outcome = 'PASSED'
            for tc in testsuite:
                if tc.good:
                    passed += 1
                else:
                    failed += 1
                    final_outcome = 'FAILED'

            summary = summary % (passed, failed)
            return export_YAML(CheckDetail(item=arg_data['item'],
                                           note=summary,
                                           report_type=arg_data['type'],
                                           outcome=final_outcome,
                                           artifact=artifact))

        else:
            assert False, "This should never happen"
