# -*- coding: utf-8 -*-
# Copyright 2009-2016, Red Hat, Inc.
# License: GPL-2.0+ <http://spdx.org/licenses/GPL-2.0+>
# See the LICENSE file for more details on Licensing

''' Utility methods related to an operating system '''

from __future__ import absolute_import
import os
import subprocess

from libtaskotron.logger import log


def is_root():
    '''Determine whether we're currently running under the root account.

    :rtype: bool
    '''
    return os.geteuid() == 0


def has_sudo():
    '''Determine whether we currently have a password-less access to sudo.

    Note: It's not possible to say whether the access will stay password-less in the future (the
    credentials might be set to expire in time), just for this exact moment.

    :rtype: bool
    '''
    # Note: We could run "sudo --reset-timestamp" first to make sure any cached password is
    # invalidated and therefore be sure we always have a password-less access. But this might be
    # seen as an impolite thing to do (esp. when we're not granted the sudo access, we would still
    # reset the cache with every run) and it might obstruct some use cases. Not implemented ATM.
    cmd = ['sudo', '--validate', '--non-interactive']
    log.debug('Deciding whether we have a password-less sudo access. Running: %s', ' '.join(cmd))

    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        log.debug(u"✘ Sudo access is not available. Received exit code %d and output:\n%s",
                  e.returncode, e.output.rstrip())
        return False
    else:
        log.debug(u'✔ Sudo access is available')
        return True
